"use strict";
/**
 * @fileoverview Interactions should be awaited
 * @author Yann Braga
 */
const utils_1 = require("../utils");
const create_storybook_rule_1 = require("../utils/create-storybook-rule");
const constants_1 = require("../utils/constants");
const ast_1 = require("../utils/ast");
module.exports = (0, create_storybook_rule_1.createStorybookRule)({
    name: '',
    defaultOptions: [],
    meta: {
        docs: {
            description: 'Interactions should be awaited',
            categories: [constants_1.CategoryId.ADDON_INTERACTIONS, constants_1.CategoryId.RECOMMENDED],
            recommended: 'error', // or 'warn'
        },
        messages: {
            interactionShouldBeAwaited: 'Interaction should be awaited: {{method}}',
            fixSuggestion: 'Add `await` to method',
        },
        type: 'problem',
        schema: [],
    },
    create(context) {
        // variables should be defined here
        //----------------------------------------------------------------------
        // Helpers
        //----------------------------------------------------------------------
        // any helper functions should go here or else delete this section
        const FUNCTIONS_TO_BE_AWAITED = [
            'expect',
            'waitFor',
            'waitForElementToBeRemoved',
            'wait',
            'waitForElement',
            'waitForDomChange',
            'userEvent',
        ];
        const getNonAwaitedCallExpressions = (body = []) => {
            return (body
                .filter((b) => {
                //@ts-ignore
                return (0, ast_1.isExpressionStatement)(b) && (0, ast_1.isCallExpression)(b.expression);
            })
                //@ts-ignore
                .map((d) => d.expression));
        };
        const getNonAwaitedInitializations = (body = []) => {
            const initializations = body
                //@ts-ignore
                .flatMap((b) => {
                return ((0, ast_1.isVariableDeclaration)(b) &&
                    b.declarations
                        .filter((d) => (0, ast_1.isVariableDeclarator)(d) && !(0, ast_1.isAwaitExpression)(d.init))
                        .map((d) => d.init));
            })
                .filter(Boolean);
            return initializations;
        };
        //@ts-ignore
        const getMethodThatShouldBeAwaited = (expression) => {
            const shouldAwait = (name) => {
                //@ts-ignore
                return FUNCTIONS_TO_BE_AWAITED.includes(name) || name.startsWith('findBy');
            };
            if ((0, ast_1.isCallExpression)(expression) &&
                (0, ast_1.isMemberExpression)(expression.callee) &&
                (0, ast_1.isIdentifier)(expression.callee.object) &&
                shouldAwait(expression.callee.object.name)) {
                return expression.callee.object;
            }
            if ((0, ast_1.isCallExpression)(expression) &&
                (0, ast_1.isMemberExpression)(expression.callee) &&
                (0, ast_1.isIdentifier)(expression.callee.property) &&
                shouldAwait(expression.callee.property.name)) {
                return expression.callee.property;
            }
            return null;
        };
        //----------------------------------------------------------------------
        // Public
        //----------------------------------------------------------------------
        /**
         * @param {import('eslint').Rule.Node} node
         */
        let invocationsThatShouldBeAwaited = [];
        return {
            AssignmentExpression(node) {
                if (!(0, ast_1.isExpressionStatement)(node.parent)) {
                    return null;
                }
                if ((0, utils_1.isPlayFunction)(node)) {
                    const { right } = node;
                    const expressionBody = (right.body && right.body.body) || [];
                    const callExpressions = [
                        ...getNonAwaitedCallExpressions(expressionBody),
                        ...getNonAwaitedInitializations(expressionBody),
                    ];
                    callExpressions.forEach((expression) => {
                        const method = getMethodThatShouldBeAwaited(expression);
                        if (method) {
                            invocationsThatShouldBeAwaited.push(method);
                        }
                    });
                }
            },
            'Program:exit': function () {
                if (invocationsThatShouldBeAwaited.length) {
                    //@ts-ignore
                    invocationsThatShouldBeAwaited.forEach((node) => {
                        context.report({
                            node,
                            messageId: 'interactionShouldBeAwaited',
                            data: {
                                method: node.name,
                            },
                            // @TODO: make this auto-fixable. Currently it's pretty dumb so something like this can happen:
                            // canvas.findByText => canvas.await findByText
                            // instead of the correct: await canvas.findByText
                            // fix: function (fixer) {
                            //   return fixer.insertTextBefore(node, 'await ')
                            // },
                            // suggest: [
                            //   {
                            //     messageId: 'fixSuggestion',
                            //     fix: function (fixer) {
                            //       return fixer.insertTextBefore(node, 'await ')
                            //     },
                            //   },
                            // ],
                        });
                    });
                }
            },
        };
    },
});
