"use strict";
/**
 * @fileoverview Use expect from '@storybook/expect'
 * @author Yann Braga
 */
const utils_1 = require("../utils");
const constants_1 = require("../utils/constants");
const ast_1 = require("../utils/ast");
const create_storybook_rule_1 = require("../utils/create-storybook-rule");
module.exports = (0, create_storybook_rule_1.createStorybookRule)({
    name: 'use-storybook-expect',
    defaultOptions: [],
    meta: {
        type: 'suggestion',
        fixable: 'code',
        hasSuggestions: true,
        schema: [],
        docs: {
            description: 'Use expect from `@storybook/jest`',
            categories: [constants_1.CategoryId.ADDON_INTERACTIONS, constants_1.CategoryId.RECOMMENDED],
            recommended: 'error',
        },
        messages: {
            updateImports: 'Update imports',
            useExpectFromStorybook: 'Do not use expect from jest directly in the story. You should use from `@storybook/jest` instead.',
        },
    },
    create(context) {
        // variables should be defined here
        //----------------------------------------------------------------------
        // Helpers
        //----------------------------------------------------------------------
        const getExpressionStatements = (body = []) => {
            return body.filter((b) => (0, ast_1.isExpressionStatement)(b));
        };
        //@ts-ignore
        const isExpect = (expression) => {
            return ((0, ast_1.isCallExpression)(expression) &&
                (0, ast_1.isMemberExpression)(expression.callee) &&
                (0, ast_1.isCallExpression)(expression.callee.object) &&
                (0, ast_1.isIdentifier)(expression.callee.object.callee) &&
                expression.callee.object.callee.name === 'expect');
        };
        const isExpectFromStorybookImported = (node) => {
            return (node.source.value === '@storybook/jest' &&
                node.specifiers.find((spec) => spec.imported.name === 'expect'));
        };
        //----------------------------------------------------------------------
        // Public
        //----------------------------------------------------------------------
        let isImportingFromStorybookExpect = false;
        let expectInvocations = [];
        return {
            ImportDeclaration(node) {
                if (isExpectFromStorybookImported(node)) {
                    isImportingFromStorybookExpect = true;
                }
            },
            AssignmentExpression(node) {
                if (!(0, ast_1.isExpressionStatement)(node.parent)) {
                    return null;
                }
                if ((0, utils_1.isPlayFunction)(node)) {
                    const { right } = node;
                    const expressionBody = (right.body && right.body.body) || [];
                    const expressionStatements = getExpressionStatements(expressionBody);
                    expressionStatements.forEach(({ expression }) => {
                        if (isExpect(expression)) {
                            expectInvocations.push(expression);
                        }
                    });
                }
            },
            'Program:exit': function () {
                if (!isImportingFromStorybookExpect && expectInvocations.length) {
                    //@ts-ignore
                    expectInvocations.forEach((node) => {
                        context.report({
                            node,
                            messageId: 'useExpectFromStorybook',
                            fix: function (fixer) {
                                return fixer.insertTextAfterRange([0, 0], "import { expect } from '@storybook/jest';\n");
                            },
                            suggest: [
                                {
                                    messageId: 'updateImports',
                                    fix: function (fixer) {
                                        return fixer.insertTextAfterRange([0, 0], "import { expect } from '@storybook/jest';\n");
                                    },
                                },
                            ],
                        });
                    });
                }
            },
        };
    },
});
